subroutine uv_merge_many(line,error)
  use gkernel_interfaces
  use gkernel_types
  use gbl_message
  !---------------------------------------------------------------------
  ! TASK  Merge Many UV tables;
  !
  ! UV_MERGE OutFile /FILES In1 In2 ... Inn /WEIGHTS W1 ... Wn /
  !   /SCALES F1 ... Fn [/MODE STACK|CONTINUUM [Index [Frequency]]
  !
  ! Merge many UV data files, with calibration factor and weight factors
  ! and spectral resampling as in the first  one.
  !
  ! With /CONTINUUM option, merge continuum UV Tables. A spectral index
  ! can then be automatically applied to the scales and weights.
  !---------------------------------------------------------------------
  character(len=*), intent(inout) :: line ! Command line
  logical, intent(out) :: error
  ! global
  integer, parameter :: o_files=1, o_mode=2, o_factor=3, o_weight=4
  character(len=*), parameter :: rname='UV_MERGE'
  ! Local
  character(len=varname_length) :: listname
  character(len=filename_length) :: table_out
  character(len=filename_length), allocatable :: table_in(:)
  type(sic_descriptor_t) :: desc
  integer(kind=address_length) :: addr
  real, allocatable :: weight(:), factor(:)
  integer :: mode, i, n, nt, ier
  character(len=4)  :: cmode
  character(len=80) :: mess
  real :: spindex   ! Spectral index
  real :: freqref   ! Reference Frequency
  logical :: err, found
  !------------------------------------------------------------------------
  ! Code:
  !
  error = .true.
  nt = sic_narg(o_files)
  !
  if (nt.lt.2) then
    ! This may be a character array handling the file names
    call sic_ch(line,o_files,1,listname,n,.true.,error)
    if (error) return
    !
    call sic_descriptor(listname,desc,found)
    if (.not.found) then
      nt = 1
    else
      !
      ! Check the Descriptor: it must be a character array
      if (desc%type.lt.1 .or. desc%ndim.ne.1) then
        call map_message(seve%e,rname,trim(listname)//' is not a character array')
        error = .true.
        return
      endif
      nt = desc%dims(1)
    endif
  endif
  !
  if (nt.lt.2) then
    call map_message(seve%e,rname,'Require at least 2 input tables')
    return
  endif
  !
  allocate(table_in(nt),weight(nt),factor(nt),stat=ier)
  if (ier.ne.0) then
    call map_message(seve%e,rname,'Memory allocation error')
    return
  endif
  weight = 1.0
  factor = 1.0
  !
  if (sic_present(o_weight,0)) then
    if (sic_narg(o_weight).ne.nt) then
      call map_message(seve%e,rname,'Number of Files and Weights mismatch')
      return
    else
      do i=1,nt
        call sic_r4(line,o_weight,i,weight(i),.true.,err)
        if (err) return
      enddo
      if (any(weight.le.0.)) then
        call map_message(seve%e,rname,'Weights must be positive')
        return
      endif
    endif
  endif
  !
  if (sic_present(o_factor,0)) then
    if (sic_narg(o_factor).ne.nt) then
      call map_message(seve%e,rname,'Number of Files and factors mismatch')
      return
    else
      do i=1,nt
        call sic_r4(line,o_factor,i,factor(i),.true.,err)
        if (err) return
      enddo
      if (any(factor.le.0.)) then
        call map_message(seve%e,rname,'factors must be positive')
        return
      endif
    endif
  endif  
  !
  if (sic_narg(o_files).eq.1) then
    ! Names from Character array
    addr = desc%addr
    do i=1,nt
      call destoc(desc%type,addr,table_in(i))
      addr = addr + desc%type
    enddo
  else
    ! Names from Command line
    do i=1,nt
      call sic_ch(line,o_files,i,table_in(i),n,.true.,err)
      if (err) return
    enddo
  endif
  error = .false.
  !
  call sic_ch(line,0,1,table_out,n,.true.,error)
  if (error) return
  !
  mode = 0  ! Simple line merge
  if (sic_present(o_mode,0)) then
    call sic_ke(line,o_mode,1,cmode,n,.false.,error)
    select case(cmode)
    case ('CONT')   ! For CONTinuum
      ! Continuum mode
      spindex = 0.
      call sic_r4(line,o_mode,2,spindex,.false.,error)
      freqref = 0.
      call sic_r4(line,o_mode,3,freqref,.false.,error)
      mode = 1
      write(mess,'(A,F6.2,A,F8.0,A)') 'Selected continuume mode, Spectral index ', &
        & spindex,' Reference frequency ',freqref,' MHz'
      call map_message(seve%i,rname,mess)
    case ('STAC')   ! For STACking  
      ! Line-Stacking mode
      call map_message(seve%e,rname,'Selected Line Stacking mode')
      mode = 2
    case default
      call map_message(seve%e,rname,'No such mode '//cmode)
      error = .true.
      return
    end select
  endif    
  !
  call sub_uv_merge_block(nt, table_in, table_out, &
    weight, factor, mode, spindex, freqref, error)
  return
!
contains

!
subroutine sub_uv_merge_block(n_in, table_in, table_out, &
     weight, factor, mode, spindex, freqref, error)
  use gildas_def
  use gkernel_interfaces
  use image_def
  use gbl_message
  !---------------------------------------------------------------------
  ! TASK  Merge an arbitray number of UV tables;
  !
  ! with calibration factor and weight factor and resampling of
  ! the first one, the second one being used as the reference
  !
  ! Input :
  ! two UV tables
  ! Output :
  ! one new UV table
  !---------------------------------------------------------------------
  integer, intent(in) :: n_in           ! Number of input tables
  character(len=*), intent(in) :: table_out   ! Output Table name
  character(len=*), intent(in) :: table_in(*) ! Other tables names
  real, intent(in)  :: weight(*)        
  real, intent(in)  :: factor(*)
  integer, intent(in)  :: mode
  logical, intent(out) :: error
  real, intent(in) :: spindex
  real, intent(in) :: freqref
  ! Local
  character(len=*), parameter :: rname='UV_MERGE'
  character(len=256) :: table_ref   ! Reference Table name
  type(gildas) :: in,inref,out
  integer, allocatable :: iwork(:,:)
  real(4), allocatable :: rwork(:,:)
  real :: scale_uv
  real(8) :: rfreq
  real :: fact, weig
  integer :: nblock, i, ier, im
  character(len=80) :: mess
  integer :: code
  integer(kind=size_length) :: in_gil_dim1,in_gil_dim2, nvisi
  integer, parameter :: code_cont=1
  integer, parameter :: code_merge=0
  integer, parameter :: code_stack=2
  !------------------------------------------------------------------------
  ! Code:
  !  !
  if (n_in.eq.0) goto 999
  if (len_trim(table_out).eq.0) goto 999
  !
  table_ref = table_in(1)
  if (len_trim(table_ref).eq.0) goto 999
  !
  ! Input file TABLE_REF  (INREF) is the reference one
  call gildas_null(inref, type = 'UVT')
  call gdf_read_gildas (inref, table_ref, '.uvt', error, data=.false.)
  if (error) then
    call map_message(seve%e,rname,'Cannot read input UV table #1 (Reference)')
    return
  endif
  !
  ! Prepare output table: compute number of visibilities
  call gildas_null(out, type = 'UVT')
  call gdf_copy_header (inref, out, error)
  call gdf_close_image (inref,error)
  call sic_parse_file(table_out,' ','.uvt',out%file)
  !
  ! Loop over tables to be merged
  call gildas_null(in, type = 'UVT')
  in_gil_dim1 = inref%gil%dim(1)
  in_gil_dim2 = inref%gil%dim(2)
  !
  if (freqref.eq.0) then
    rfreq = inref%gil%freq
  else
    rfreq = freqref
  endif
  !
  do im=2,n_in
    call gdf_read_gildas (in, table_in(im), '.uvt', error, data=.false.)
    if (error) then
      write(mess,*) 'Cannot read input UV table #',im
      call map_message(seve%e,rname,mess)
      return
    endif
    call gdf_close_image(in,error)
    !
    if (mode.eq.code_cont) then
      if (in%gil%nchan.ne.1) then
        call map_message(seve%e,rname,'CONTINUUM mode is only valid for 1 channel')
        error = .true.
        return
      endif
    endif
    !
    out%gil%dim(2) = in%gil%dim(2) + out%gil%dim(2)
    out%gil%nvisi = out%gil%dim(2)  ! Set the number of visibilities
    in_gil_dim1 = max(in_gil_dim1,in%gil%dim(1))
    in_gil_dim2 = max(in_gil_dim2,in%gil%dim(2))
    !
    ! Verify if trailing columns do match
    call gdf_uvmatch_codes(inref,in,code)
    if (code.eq.3) then
      ! Column # 3 has a mismatch - Make sure the Ref is now SCAN
      inref%gil%column_pointer(code_uvt_w) = 0
      inref%gil%column_size(code_uvt_w) = 0
      inref%gil%column_pointer(code_uvt_scan) = 3
      inref%gil%column_size(code_uvt_scan) = 1
     else if (code.lt.0) then
      error = .true. 
      call map_message(seve%e,rname,'Leading or Trailing columns do not match')
      return
    endif
    !
  enddo
  !
  ! Finished preparatory work, and create the OUT table
  call map_message(seve%i,rname,'Creating UV table '//trim(out%file))
  call gdf_create_image(out, error)
  if (error) then
    call map_message(seve%e,rname,'Cannot create output UV table')
    return
  endif
  !
  ! Define blocking factor,
  call gdf_nitems('SPACE_GILDAS',nblock,inref%gil%dim(1)) ! Visibilities at once
  nblock = min(nblock,inref%gil%dim(2))
  allocate (out%r2d(out%gil%dim(1),nblock), stat=ier)
  if (ier.ne.0) then
    write(mess,*) 'Memory allocation error ',inref%gil%dim(1), nblock
    call map_message(seve%e,rname,mess)
    error = .true.
    return
  endif
  !
  ! Define blocking factor
  call gdf_nitems('SPACE_GILDAS',nblock,in_gil_dim1) ! Visibilities at once
  nblock = min(nblock,in_gil_dim2)
  allocate (out%r2d(out%gil%dim(1),nblock), stat=ier)
  if (ier.ne.0) then
    write(mess,*) 'Memory allocation error OUT ',out%gil%dim(1), nblock
    call map_message(seve%e,rname,mess)
    error = .true.
    return
  endif
  !
  if (out%gil%nchan.eq.1 .and. in%gil%nchan.eq.1) then
    call map_message(seve%i,rname,'Merging continuum tables')
  else
    allocate (iwork(2,out%gil%nchan), rwork(4,out%gil%nchan), stat=ier)
    if (ier.ne.0) then
      write(mess,*) 'Memory allocation error WORK ',6,out%gil%nchan
      call map_message(seve%e,rname,mess)
      error = .true.
      return
    endif
    call map_message(seve%i,rname,'Merging line tables')
  endif
  !
  ! Append INs to OUT with resampling and rescaling when needed
  out%blc(2) = 0    ! Get started
  !
  do im = 1,n_in
    mess = 'Reading '//trim(table_in(im))
    call map_message(seve%i,rname,mess)
    call gdf_read_gildas (in, table_in(im), '.uvt', error, data=.false.)
    if (error) then
      write(mess,*) 'Cannot read input UV table #',im
      call map_message(seve%e,rname,mess)
      return
    endif
    !
    allocate (in%r2d(in%gil%dim(1),nblock), stat=ier)
    if (ier.ne.0) then
      write(mess,*) 'Memory allocation error IN ',in%gil%dim(1), nblock
      call map_message(seve%e,rname,mess)
      error = .true.
      return
    endif
    !
    if (out%gil%nchan.eq.1 .and. in%gil%nchan.eq.1) then
      scale_uv = in%gil%freq / rfreq
    else
      if (mode.eq.code_stack) then
        !   In Stacking Mode, Lambda/D is conserved, so the UV length
        ! must be scaled by FreqIn/ FreqRef
        scale_uv = in%gil%freq / inref%gil%freq
      else
        scale_uv = 1.0
      endif
    endif
    fact = factor(im)
    weig = weight(im)
    if (out%gil%nchan.eq.1 .and. in%gil%nchan.eq.1) then
      if (spindex.ne.0) then
        fact = factor(im) / (scale_uv)**spindex
        weig = weight(im) * (scale_uv)**(2*spindex)
      endif
    endif    
    write(mess,'(A,F8.2,A,F8.2,A,F8.4)') 'Factors: Flux ',fact, &
      & ', Weight ',weig,' UV ',scale_uv
    call map_message(seve%i,rname,mess)
    !
    out%blc(2) = out%trc(2)+1
    !
    in%blc = 0
    in%trc = 0
    !
    ! Block loop on current input file
    do i=1,in%gil%dim(2),nblock
      write(mess,*) i,' / ',in%gil%dim(2),nblock
      call map_message(seve%d,rname,mess)
      in%blc(2) = i
      in%trc(2) = min(in%gil%dim(2),i-1+nblock)
      nvisi = in%trc(2)-in%blc(2)+1
      out%trc(2) = out%blc(2)+nvisi-1
      call gdf_read_data(in,in%r2d,error)
      !
      if (out%gil%nchan.eq.1 .and. in%gil%nchan.eq.1) then
        call uvmergec (nvisi, out%gil%dim(1), out%r2d,  &
       &       in%gil%dim(1),in%r2d,weig,fact,      &
       &       scale_uv)
      else
        call uvmergel (nvisi, out%gil%nchan, out%r2d,    &
       &       in%gil%nchan,in%r2d,weig,fact, scale_uv,  &
       &       iwork, rwork, &
       &       out%gil%ref(1), dble(out%gil%voff), dble(out%gil%vres), &
       &       in%gil%ref(1), dble(in%gil%voff), dble(in%gil%vres) )
      endif
      call gdf_write_data(out,out%r2d,error)
      if (error) return
      out%blc(2) = out%trc(2)+1
    enddo
    !
    call gdf_close_image(in,error)
    deallocate (in%r2d,stat=ier)
  enddo
  !
  call gdf_close_image(out,error)
  deallocate(out%r2d)
  call map_message(seve%i,rname,'Successful completion')
  return
  !
  999   error = .true.
end subroutine sub_uv_merge_block
!
subroutine gdf_uvmatch_codes(a,b,code)
  use image_def
  ! use gio_interfaces, only : gdf_uvcolumn_codes
  !---------------------------------------------------------------------
  ! @ public
  !
  ! GIO     Compare layout of UV tables for description columns
  ! Code = 0    Same layout
  ! Code = -1   Different layout
  ! Code = 3    Similar layout, except that Col.3 is SCAN in one,
  !             and W in the other.
  !---------------------------------------------------------------------
  type(gildas), intent(in) :: a       ! Input Header
  type(gildas), intent(in) :: b       ! Input Header
  integer, intent(out)     :: code    ! Returned code
  !
  integer :: k, nlt, ier
  integer, allocatable :: a_codes(:,:), b_codes(:,:)
  !
  ! If # number of trailing : no match
  code = -1
  if (a%gil%ntrail.ne.b%gil%ntrail) return
  ! If # number of leading : no match
  if (a%gil%nlead.ne.b%gil%nlead) return
  !
  ! Else, compare columns, but make a specific provision
  ! for SCAN and W (both of which may happen in Col 3)
  !
  nlt = a%gil%nlead + a%gil%ntrail
  allocate(a_codes(2,nlt),b_codes(2,nlt),stat=ier)
  call gdf_uvcolumn_codes(a,a_codes)
  call gdf_uvcolumn_codes(b,b_codes)  
  !
  code = 0
  do k = 1,nlt
    if (a_codes(1,k).ne.b_codes(1,k)) then
      if (code.eq.0 .and. k.eq.3) then
        code = 3
        if (a_codes(1,k).eq.code_uvt_scan .and. b_codes(1,k).eq.code_uvt_w) cycle 
        if (b_codes(1,k).eq.code_uvt_scan .and. a_codes(1,k).eq.code_uvt_w) cycle
      endif
      code = -1
    else if (a_codes(2,k).ne.b_codes(2,k)) then
      Print *,'Code length mismatch at Col. ',k
    endif
  enddo
end subroutine gdf_uvmatch_codes
!
subroutine gdf_uvcolumn_codes(h,column_codes)
  use image_def
  !---------------------------------------------------------------------
  ! @ public-mandatory
  !
  ! GIO     Return the column codes and sizes of the Leading and Trailing
  !         columns
  !---------------------------------------------------------------------
  type (gildas), intent(in) :: h                ! Input header
  integer, intent(inout) :: column_codes(:,:)   ! Column codes
  !
  integer :: nlt, i, j, itrail
  !
  nlt = h%gil%nlead + h%gil%ntrail
  !
  do j = 1,nlt
    do i = 1, code_uvt_last
      if (h%gil%column_pointer(i).lt.h%gil%fcol) then
        if (h%gil%column_pointer(i).eq.j) then
          column_codes(1,j) = i
          column_codes(2,j) = h%gil%column_size(i)
          exit
        endif
      else if (h%gil%column_pointer(i).gt.h%gil%lcol) then
         ! A little more vicious if > nlead
        itrail = h%gil%column_pointer(i)-h%gil%lcol+h%gil%nlead
        !! Print *,'Trail ',j,i,itrail
        if (itrail.eq.j) then
          column_codes(1,j) = i
          column_codes(2,j) = h%gil%column_size(i)
          exit
        endif
      else
        Print *,'Inconsistent Column Pointer #',i,' = ', h%gil%column_pointer(i), &
        &   ' in data range',h%gil%fcol,h%gil%lcol
      endif
    enddo
  enddo
end subroutine gdf_uvcolumn_codes
!
subroutine uvmergel (nv,ncx,x,ncz,z,weight,factor,scaleuv,iwork,rwork &
  & ,xref,xval,xinc,zref,zval,zinc)
  use gildas_def
  !------------------------------------------------------------------------
  ! Merge two line data sets:  Append them after interpolation (if needed)
  !------------------------------------------------------------------------
  integer(kind=size_length), intent(in) :: nv   ! Number of visibilities
  integer, intent(in) :: ncx                    ! Number of channels
  real, intent(inout) :: x(:,:)                 ! Merged UV table
  integer, intent(in) :: ncz                    ! Number of channels
  real, intent(in) :: z(:,:)                    ! Input UV table
  real, intent(in) :: weight                    ! Weight scale factor
  real, intent(in) :: factor                    ! Intensity scale factor
  real, intent(in) :: scaleuv                   ! UV scale factor
  integer, intent(inout) :: iwork(2,*)          ! Work array
  real, intent(inout) :: rwork(4,*)             ! Work array
  real(8), intent(in) :: xref,xval,xinc         ! Input velocity sampling
  real(8), intent(in) :: zref,zval,zinc         ! Output velocity sampling
  !------------------------------------------------------------------------
  !
  ! Local
  integer :: i, j, k, iv
  real :: poids
  !
  ! iwork = imin,imax
  ! rwork = rmin_1, rmin, rmax, rmax_1
  !
  integer imax,imin, ix,iy, jxmin, jymin, jxmax, jymax
  integer :: mcx, mcz, lcx, lcz, mtrail
  real(8) minpix, maxpix, pix, val, expand
  real rmin_1, rmin, rmax_1, rmax, smin, smax
  real scale
  !------------------------------------------------------------------------
  ! Code:
  ! Interpolate first (Z) Table on X grid
  !
  mcx = ubound(x,1)
  mcz = ubound(z,1)
  lcx = 3*ncx+7
  lcz = 3*ncz+7
  mtrail = mcx-lcx
  ! Setup interpolation scheme
  !
  expand = abs(xinc/zinc)
  scale  = factor/expand
  !
  ! Initialize workspace
  iwork(:,1:ncx) = 0
  rwork(:,1:ncx) = 0.0
  do i = 1, ncx
    !
    ! Compute interval
    val = xval + (i-xref)*xinc
    pix = (val-zval)/zinc + zref
    maxpix = pix + 0.5d0*expand
    minpix = pix - 0.5d0*expand
    imin = int(minpix+1.0d0)
    imax = int(maxpix)
    iwork(1,i) = imin
    iwork(2,i) = imax
    if (imax.eq.imin .and. expand.eq.1.0) then
      ! Shifted grid or not actually
      rwork(1,i) = (imin-pix)
      rwork(2,i) = 1.0-rwork(1,i)
    else if (imax.ge.imin) then
      ! Lower end
      if (imin.gt.1) then
        !
        ! ZMIN = Z(IMIN-1)*(IMIN-MINPIX)+Z(IMIN)*(MINPIX-IMIN+1)
        !     and
        ! X(I) = X(I) + 0.5 * ( (IMIN-MINPIX)*(ZMIN+Z(IMIN)) )
        !
        !     RMIN_1 = 0.5 * (IMIN-MINPIX)*(IMIN-MINPIX)     ! Coefficient de IMIN-1
        !     RMIN   = 0.5 * (MINPIX-IMIN+2)*(IMIN-MINPIX)   ! Coefficient de IMIN
        !
        smin = imin-minpix
        rwork(1,i) = 0.5*smin*smin
        rwork(2,i) = smin-rwork(1,i)
      else
        !     RMIN   = (IMIN-MINPIX)                          ! Coefficient de IMIN
        !     RMIN_1 = 0                                      ! Coefficient de IMIN-1
        rwork(1,i) = 0.0
        rwork(2,i) = imin-minpix
      endif
      ! Upper end
      if (imax.lt.ncz) then
        smax = maxpix-imax
        rwork(4,i) = 0.5*smax*smax
        rwork(3,i) = smax-rwork(4,i)
      else
        rwork(3,i) = maxpix-imax
        rwork(4,i) = 0.0
      endif
      if (imax.eq.imin) then
        ! 3 Point Case
        rwork(2,i) = rwork(2,i)+rwork(3,i)
        rwork(3,i) = 0.0
      else
        ! General Case: Add 1/2 of IMIN & IMAX + 1 of all intermediate channels
        rwork(2,i) = rwork(2,i)+0.5
        rwork(3,i) = rwork(3,i)+0.5
      endif
    else
      if (imin.gt.1) then
        rwork(1,i) = (imin-pix)
        rwork(2,i) = 1.0-rwork(1,i)
      else
        rwork(1,i) = 0.0
        rwork(2,i) = 1.0
      endif
    endif
  enddo
  !
  ! OK, now do it for all visibilities
  do iv=1,nv
    !
    ! Coordinates
    do j=1, 3
      x(j,iv) = z(j,iv)*scaleuv
    enddo
    do j=4, 7
      x(j,iv) = z(j,iv)
    enddo
    !
    ! Channels
    do i=1,ncx
      imin = iwork(1,i)
      imax = iwork(2,i)
      jxmin = 5+3*imin
      jymin = jxmin+1
      jxmax = 5+3*imax
      jymax = jxmax+1
      rmin_1 = rwork(1,i)
      rmin   = rwork(2,i)
      rmax   = rwork(3,i)
      rmax_1 = rwork(4,i)
      ix = 5+3*i
      iy = ix+1
      !
      if (imax.ge.imin) then
        ! General Trapezium integration
        if (imin.gt.1) then
          x(ix,iv) = z(jxmin-3,iv)*rmin_1+z(jxmin,iv)*rmin
          x(iy,iv) = z(jymin-3,iv)*rmin_1+z(jymin,iv)*rmin
        else
          x(ix,iv) = z(jxmin,iv)*rmin
          x(iy,iv) = z(jymin,iv)*rmin
        endif
        if (imax.gt.imin) then
          do k = 5+3*(imin+1),5+3*(imax-1),3
            x(ix,iv) = x(ix,iv) + z(k,iv)
            x(iy,iv) = x(iy,iv) + z(k+1,iv)
          enddo
        endif
        x(ix,iv) = x(ix,iv) + z(jxmax,iv)*rmax
        x(iy,iv) = x(iy,iv) + z(jymax,iv)*rmax
        if (imax.lt.ncz) then
          x(ix,iv) = x(ix,iv) + z(jxmax+3,iv)*rmax_1
          x(iy,iv) = x(iy,iv) + z(jymax+3,iv)*rmax_1
        endif
        x(ix,iv) = x(ix,iv)*scale
        x(iy,iv) = x(iy,iv)*scale
      else
        !
        ! Interpolation
        if (imin.gt.1) then
          x(ix,iv) = z(jxmin-3,iv)*rmin_1+z(jxmin,iv)*rmin
          x(iy,iv) = z(jymin-3,iv)*rmin_1+z(jymin,iv)*rmin
        else
          x(ix,iv) = z(jxmin,iv)
          x(iy,iv) = z(jymin,iv)
        endif
        x(ix,iv) = x(ix,iv)*factor
        x(iy,iv) = x(iy,iv)*factor
      endif
    enddo
    !
    ! Weight
    ! Same weight for all channels.. One should do better here.
    !
    poids = z(10+3*(ncz/2),iv)*weight
    if (poids.eq.0) then  ! May happen with Widex spectra
      poids = z(10+3*(ncz/3),iv)*weight
    endif
    do j=10, 3*ncx+7, 3
      x(j,iv) = poids
    enddo
    !
    ! Trailing columns if any
    do j=1,mtrail
      x(lcx+j,iv) = z(lcz+j,iv)
    enddo
  enddo
end subroutine uvmergel
!
subroutine uvmergec (nv,ndx,x,ndz,z,weight,factor,scaleuv)
  use gildas_def
  !------------------------------------------------------------------------
  !
  ! Append a Continuum Data set
  !------------------------------------------------------------------------
  integer(kind=size_length), intent(in) :: nv   ! Number of visibilities
  integer(kind=size_length), intent(in) :: ndx  !                  !
  real, intent(inout) :: x(ndx,nv)              ! Merged UV table
  integer(kind=size_length), intent(in) :: ndz  !
  real, intent(in) :: z(ndz,nv)                 ! Input UV table
  real, intent(in) :: weight                    ! Weight scale factor
  real, intent(in) :: factor                    ! Intensity scale factor
  real, intent(in) :: scaleuv                   ! Baseline length scale
  ! Local
  integer :: j, iv, n
  !------------------------------------------------------------------------
  ! Code:
  n = min(ndx,ndz)
  !
  ! OK, now do it for all visibilities
  do iv=1,nv
    !
    ! Coordinates
    x(1,iv) = scaleuv*z(1,iv)
    x(2,iv) = scaleuv*z(2,iv)
    do j=3, 7
      x(j,iv) = z(j,iv)
    enddo
    !
    ! Visibilities
    x(8,iv) = z(8,iv)*factor
    x(9,iv) = z(9,iv)*factor
    ! Weight
    x(10,iv) = z(10,iv)*weight
    ! Other (if any)
    do j=11,n
      x(j,iv) = z(j,iv)
    enddo
  enddo
end subroutine uvmergec
!
subroutine uvmergef (nv,ndx,x,ndz,z,weight,factor,w1,w2)
  use gildas_def
  !------------------------------------------------------------------------
  !
  ! Merge two Continuum Data Set: Average visibilities
  !
  !------------------------------------------------------------------------
  integer(kind=size_length), intent(in) :: nv   ! Number of visibilities
  integer(kind=size_length), intent(in) :: ndx  !
  real, intent(inout) :: x(ndx,nv)              ! Merged UV table
  integer(kind=size_length), intent(in) :: ndz  !
  real, intent(in) :: z(ndz,nv)                 ! Input UV table
  real, intent(in) :: weight                    ! Weight scale factor
  real, intent(in) :: factor                    ! Intensity scale factor
  real(8), intent(out) :: w1,w2                 ! Relative weights
  ! Local
  integer :: iv, n, j
  real :: ww
  !------------------------------------------------------------------------
  ! Code:
  n = min(ndx,ndz)
  w1 = 0.d0
  w2 = 0.d0
  ! OK, now do it for all visibilities
  do iv=1,nv
    !
    ! U and V Coordinates
    if (x(1,iv).ne.z(1,iv)) then
      print *,'U iv ',iv,x(1,iv),z(1,iv), x(10,iv), z(10,iv)
      print *,'Out ',z(1:10,iv)
      goto 99
    endif
    if (x(2,iv).ne.z(2,iv)) then
      print *,'V iv ',iv,x(2,iv),z(2,iv), x(10,iv), z(10,iv)
      goto 99
    endif
    !
    ! Visibilities
    ww = x(10,iv)
    if (x(10,iv).gt.0) w1 = w1+x(10,iv)
    !
    if (z(10,iv).gt.0) then
      w2 = w2+z(10,iv)
      if (x(10,iv).gt.0) then
        ww = x(10,iv)+weight*z(10,iv)
        x(8,iv) = (x(8,iv)*x(10,iv) + z(8,iv)*z(10,iv)*weight*factor ) / ww
        x(9,iv) = (x(9,iv)*x(10,iv) + z(9,iv)*z(10,iv)*weight*factor ) / ww
      else
        ww = weight*z(10,iv)
        x(8,iv) = z(8,iv)*factor
        x(9,iv) = z(9,iv)*factor
      endif
    endif
    ! Weight
    x(10,iv) = ww
    ! Trailing columns if any
    do j=11,n
      x(j,iv) = z(j,iv)
    enddo
  enddo
  return
99  call gagout('E-UV_MERGE,  Continuum tables do not match')
end subroutine uvmergef
!
end subroutine uv_merge_many
